"""Please store all the constants here."""

VIEW_MODELS = '/services/api/v1/models'
VIEW_EQUIPMENT = '/services/api/v1/equipment'
VIEW_NOTIFICATIONS = '/services/api/v1/notification'
VIEW_OBJECTS = '/services/api/v1/objects/'
VIEW_MODEL_INDICATORS = '/services/api/v1/model'
VIEW_FAILUREMODES = '/services/api/v1/failuremodes'
VIEW_LOCATIONS = '/services/api/v1/location'
VIEW_WORKORDERS = '/services/api/v1/workorders'
VIEW_SYSTEMS = '/services/api/v1/systems'
